# Copyright (C) 2008, 2009 by Intevation GmbH
#
# Based on code from
# http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/456195 which
# is Copyright (C) 2006 by Alessandro Budai
#
# This software may be used and distributed according to the terms
# of the Python License 2.3 or newer, see http://www.python.org/license

"""
Functions to establish network connections of various kinds.
In particular, the functions allow vanilla TCP connections and SSL
connections and HTTP CONNECT based connections combined in a more or
less arbitrary ways.

In addition there's a httplib.HTTPConnection sub-class that can be
instantiated with an existing socket object so that one doesn't have to
override any methods to handle special situations such as connections
via an HTTPS proxy using the HTTP CONNECT method.
"""

import httplib
import socket

# the ssl modules was introduced in Python 2.6.  If available, we use
# that, otherwise the older ssl support from the socket module
try:
    import ssl
except ImportError:
    ssl = None

def log_debug(template, *args):
    pass

def connect_tcp(host, port, log_debug=log_debug):
    """Establish a TCP connection to port port on host.
    The return value is the socket object for the connection
    """
    msg = "getaddrinfo returns an empty list"
    for res in socket.getaddrinfo(host, port, 0, socket.SOCK_STREAM):
        af, socktype, proto, canonname, sa = res
        try:
            sock = socket.socket(af, socktype, proto)
            log_debug("connect: (%s, %s)", host, port)
            sock.connect(sa)
        except socket.error, msg:
            log_debug("connect fail:", host, port)
            if sock:
                sock.close()
            sock = None
            continue
        break
    if not sock:
        raise socket.error, msg

    return sock


def connect_http_connect(sock, real_host, real_port, headers, debuglevel=0):
    """Establish a connection through a HTTP proxy using the CONNECT method.

    The function assumes sock is a socket-like object already connected
    to the proxy by whatever means is suitable for the caller.  The
    function sends the CONNECT request for the given real_host and
    real_port, and ready the response.  If the response indicates
    success, the function returns sock.  The headers parameter should be
    a list of additional sending any given additional HTTP request
    header fields.
    """
    sock.send("CONNECT %s:%d HTTP/1.0\r\n" % (real_host, real_port))
    for item in headers:
        sock.send("%s: %s\r\n" % item)
    sock.send("\r\n")
    response = httplib.HTTPResponse(sock, method="CONNECT",
                                    debuglevel=debuglevel)
    (version, code, message) = response._read_status()
    if code != 200:
        close()
        raise socket.error, ("Proxy connection failed: %d %s"
                             % (code, message.strip()))
    while True:
        line = response.fp.readline()
        if not line:
            raise socket.error, ("Proxy connection failed:"
                                 " unexpected end of response header")

        if line == '\r\n':
            break

    return sock


def connect_ssl(sock, keyfile=None, certfile=None, log_debug=log_debug):
    """Establish an SSL connection on the socket-like object sock.
    If successful, the function returns a new socket like object that
    uses SSL to encrypt the data.
    """
    log_debug("initiate ssl on %r", sock)
    if ssl is not None:
        wrapped_sock = ssl.wrap_socket(sock, keyfile, certfile)
    else:
        sslsock = socket.ssl(sock, keyfile, certfile)
        wrapped_sock = httplib.FakeSocket(sock, sslsock)
    log_debug("ssl initiated on %r", sock)
    return wrapped_sock


class SocketHTTPConnection(httplib.HTTPConnection):

    """Extends HTTPConnection to accept a predefined socket-like object.
    The base class tries to establish the connection itself, which is
    awkward given all the different ways a connection can be established
    in inteproxy, such as directly per tcp, or SSL over an HTTPs proxy.
    """

    def __init__(self, sock, host, port=None, *args, **kw):
        httplib.HTTPConnection.__init__(self, host, port, *args, **kw)
        self.sock = sock

    def connect(self):
        pass
