# Copyright (C) 2007, 2008, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Classes to modify HTTP requests."""

import re
import base64
import urlparse
from urllib import quote_plus, splitport
from StringIO import StringIO

from lxml import etree

import getpassword


class IdentityTranscoder(object):

    """A transcoder that does not change anything

    Other transcoders can be derived from this class so that they only
    need to implement the methods that actually change parts of a
    request.
    """

    def __init__(self, method, spliturl, rule):
        """Initializes the transcoder

        Parameters:
        method -- The HTTP method, either 'GET' or 'POST'
        spliturl -- A tuple with the split url of the real remote host.
                    The tuple has the form of the return value of
                    urlparse.urlsplit
        """
        self.method = method
        self.spliturl = spliturl
        self.rule = rule

    def get_password(self):
        """Returns the username and password needed for the request

        The default implementation here simply uses the host and path
        parts of the url as the parameter for the
        get_password_with_cache function.
        """
        if self.rule.user is not None:
            return (self.rule.user, self.rule.password)
        return getpassword.get_password_with_cache("".join(self.spliturl[1:3]))

    def get_url(self):
        """Returns the real url to connect to when handling this request

        The default implementation simply returns the recombined url
        given to the constructor
        """
        return urlparse.urlunsplit(self.spliturl)

    def convert_request(self, request):
        """Converts the TranscoderRequest instance.
        Derived classes should override this method to e.g. add or
        modify headers and/or the body according to its needs.
        """


class HTTPSTranscoder(IdentityTranscoder):

    """Transcoder that turns the URL unconditionally into a HTTPS URL"""

    def __init__(self, method, spliturl, rule):
        super(HTTPSTranscoder, self).__init__(method, ("https",) + spliturl[1:],
                                              rule)


class OWSProxyGETTranscoder(HTTPSTranscoder):

    """Transcoder for GET requests to the OWSProxy"""

    def get_url(self):
        """Returns the URL with username and password added to the query"""
        scheme, netloc, path, query, fragment = self.spliturl
        if self.method == "GET":
            user, password = self.get_password()
            # FIXME: what do we do if the user cancelled?
            if user is not None:
                if query:
                    split_query = query.split("&")
                else:
                    split_query = []
                query = "&".join(split_query
                                 + ["user=%s" % quote_plus(user),
                                    "password=%s" % quote_plus(password)])
        return urlparse.urlunsplit((scheme, netloc, path, query, fragment))

    def convert_request(self, request):
        """Modifies the request to avoid a timeout problem with OWSProxy.
        OWSProxy always tries to read the body of a request, regardless
        of the HTTP method and the headers of the request.  This doesn't
        work correctly if the request header doesn't indicate that the
        request has a body (with the Content-Length or Transfer-Encoding
        header fields) and the header contains a 'Connection: close'
        field.  In this case, the Java servlet API tries to read all
        data the client sends until it closes the connection (presumably
        by closing its socket for writing with shutdown).  This
        incorrect behavior according to RFC2616, section 4.4, which
        allows only the server to indicate the length of the body by
        closing the connection.

        To work around this problem, this method adds a Content-Length
        header to the request that indicates a body of length 0.  It
        doesn't add a real empty body with request.set_body() because
        that would make urllib2 turn the request into a POST request.

        Other workarounds would be to either not send the 'Connection:
        close' header field, or to shutdown the socket for writing after
        sending the request.  Both are not easy to implement with
        urllib2, though.
        """
        if request.body is None:
            request.headers["Content-Length"] = "0"


class OWSProxyPOSTTranscoder(HTTPSTranscoder):

    """Transcoder for POST requests to the OWSProxy"""

    def convert_request(self, request):
        """Adds user and password attributes to the GetFeature element.
        """
        t = etree.parse(StringIO(request.body))
        root = t.getroot()
        if root.tag in ("{http://www.opengis.net/wfs}GetFeature",
                        "{http://www.opengis.net/wfs}Transaction"):
            user, password = self.get_password()
            root.set("user", user)
            root.set("password", password)
            request.set_body(etree.tostring(t, xml_declaration=True),
                             "text/xml")


class BasicAuthTranscoder(HTTPSTranscoder):

    """Transcoder for requests that use HTTP Basic Authentication"""

    def convert_request(self, request):
        """Asks the user for credentials and adds a Basic Authorization header
        """
        user, password = self.get_password()
        if user is not None:
            basicauth = "Basic " + base64.b64encode(user + ":" + password)
            request.headers["Authorization"] = basicauth


def pattern_to_regex(pattern, character_set="."):
    """Returns a regular expression string for a wild-card pattern.
    The wild-card character is '*'.  The character_set argument should
    be a regular expression matching exactly one of the characters
    matched by the wild-card character.  The default value of
    character_set is '.'
    """
    wildcard_regex = character_set + "*"
    return wildcard_regex.join(re.escape(part) for part in pattern.split("*"))


class TranscoderMap(object):

    """Manages the host specific transcoder settings"""

    def __init__(self, classes):
        """Initializes the TranscoderMap.

        The parameter 'classes' should be a sequence of (name, get,
        post) triples.  They will be used to add classes with the
        add_class method.  The first item in classes will become the
        default class.
        """
        self.rules = []
        self.classmap = dict()
        self.default_classname = None

        if classes:
            for item in classes:
                self.add_class(*item)
            self.set_default_class(classes[0][0])

    def set_default_class(self, name):
        """Sets the default class"""
        self.default_classname = name

    def add_class(self, name, get, post):
        """Adds a transcoder class named name.
        The get and post parameters are the python classes implementing
        the transcoder for the GET and POST methods.
        """
        self.classmap[name] = dict(GET=get, POST=post)

    def add_rule(self, rule):
        """Adds a transcoder rule (typically an instance of InteProxyRule)
        """
        netloc = rule.host
        if rule.port is not None:
            netloc += ":%d" % rule.port
        # characters that may not occur in URIs.  See RFC 2396, section 2.4.3
        # We do not exclude '%' because that's used to escape characters
        # in URLs, so it's likely to occur in the actual textual
        # representation of the URLs we are working with.
        excluded_chars = "\x00-\x20\x7F<>#\""
        re_netloc = pattern_to_regex(netloc,
                                     character_set=r"[^/%s]" % excluded_chars)
        re_path = pattern_to_regex(rule.path,
                                   character_set=r"[^%s]" % excluded_chars)
        self.rules.append((re.compile(re_netloc), re.compile(re_path), rule))

    def add_rules(self, rules):
        """Adds all transcoder rules in rules"""
        for rule in rules:
            self.add_rule(rule)

    def lookup(self, netloc, path):
        """Returns the rule matching host and path.
        If no rule matches, the method returns None
        """
        for netloc_regex, path_regex, rule in self.rules:
            netloc_match = netloc_regex.match(netloc)
            if netloc_match and netloc_match.group(0) == netloc:
                path_match = path_regex.match(path)
                if path_match and path_match.group(0) == path:
                    return rule
        return None

    def get_transcoder(self, method, url):
        """Returns a transcoder for the given HTTP method and URL

        URL should be the URL given for the GET or POST http request.
        This function extracts the real remote host from the URL and
        uses the URL to look up the transcoder with lookup_transcoder.

        This method supports two ways to encode the remote host in the
        URL:

           -- InteProxy is accessed directly as an HTTP server.  In this
              cases URL is an absolute pathname and does not include a
              hostname in the normal URL syntax.  Instead get_transcoder
              assumes that it encodes the real remote host in the first
              component of the path.  This is the old InteProxy method

           -- InteProxy is used as a HTTP-proxy.  In this case URL will
              be a full URL including a hostname.  In this case
              get_transcoder will use the host of the URL as the real
              remote host.  The path component of the URL is not
              modified.
        """
        scheme, netloc, path, query, fragment = urlparse.urlsplit(url)

        # If scheme is given, InteProxy is used as a HTTP proxy, otherwise
        # it's the old InteProxy approach.  Only the old approach needs
        # special casing.
        if not scheme:
            scheme = "http"
            # determine the remote host encoded in the path
            split_path = path.split("/")
            if not split_path[0]:
                del split_path[0]
            netloc = split_path.pop(0)
            path = "/" + "/".join(split_path)

        inteproxy_scheme = self.default_classname
        rule = self.lookup(netloc, path)
        if rule is not None:
            inteproxy_scheme = rule.scheme
        transcoder_class = self.classmap[inteproxy_scheme][method]
        return transcoder_class(method,
                                (scheme, netloc, path, query, fragment),
                                rule)

    def rewrite_urls(self, data, prefix, log_debug):
        """Prefix all known URLs in data with prefix.

        This method looks for occurrences of all urls that this
        transcode map is responsible for in data replaces them with
        equivalent urls that refer to this inteproxy instance explicitly
        by prefixing them with the given prefix.  The prefix should be
        the URL of the InteProxy instance.  E.g. if the proxy is
        responsible for connections to http://example.com/wms, and this
        URL occurs in the body, it will be replaced by
        http://localhost:64609/example.com/wms (this example assumes the
        default port for inteproxy).

        The log_debug parameter should be a callable for debug log
        messages, usually the log_debug method of the
        InteProxyHTTPRequestHandler.
        """
        url_patterns = []
        for host_regex, path_regex, classname in self.rules:
            url_patterns.append("%s%s"
                                % (host_regex.pattern, path_regex.pattern))
        regex = ("(?:http|https)://(?:"
                 + "|".join("(" + pattern + ")"
                            for pattern in url_patterns)
                 + ")")

        def make_inteprox_url(match):
            url = match.group(match.lastindex)
            return prefix + url

        return re.sub(regex, make_inteprox_url, data)


def create_transcoder_map():
    return TranscoderMap([
        ("identity", IdentityTranscoder, IdentityTranscoder),
        ("owsproxy", OWSProxyGETTranscoder, OWSProxyPOSTTranscoder),
        ("basicauth", BasicAuthTranscoder, BasicAuthTranscoder),
        ])
