# Copyright (C) 2007, 2008, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Simple HTTP server with some static data that can be run in a thread"""


import threading
import socket
import urllib
import base64
import BaseHTTPServer

try:
    import ssl
except ImportError:
    ssl = None

import inteproxy.httpserver as httpserver
from inteproxy.httpconnection import connect_tcp

class HTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    def do_GET(self):
        """Handle a GET request from the server's contents attribute.
        See HTTPServer for more details.
        """
        for path, headers, data in self.server.contents:
            if path == self.path:
                self.send_response(200)
                for header, value in headers:
                    self.send_header(header, value)
                self.send_header("Content-Length", len(data))
                self.end_headers()
                self.wfile.write(data)
                return
        else:
            self.send_error(404)


class HTTPServer(httpserver.HTTPServer):

    """HTTPServer that serves a limited set of predefined contents.

    The contents are described by a list of tuples of the form (PATH,
    HEADERS, DATA), where PATH and DATA are strings describing the path
    on the server and the data found at that path.  HEADERS should be a
    list of (HEADER-KEYWORD, VALUE) pairs describing the HTTP-header
    fields of the resource at PATH.  The Content-Length header field is
    automatically added by the request handler.
    """

    def __init__(self, contents, server_address=("127.0.0.1", 0),
                 RequestHandlerClass=HTTPRequestHandler):
        self.contents = contents
        httpserver.HTTPServer.__init__(self, server_address,
                                       RequestHandlerClass)


def start_copy_thread(insock, outsock):
    def copy():
        # copy some constants to local scope to avoid attribute errors
        # that may occur at the end of the test runs when the
        # interpreter shuts down
        SHUT_WR = socket.SHUT_WR
        SHUT_RD = socket.SHUT_RD
        socket_error = socket.error
        while True:
            chunk = insock.recv(1024)
            if not chunk:
                outsock.shutdown(SHUT_WR)
                try:
                    insock.shutdown(SHUT_RD)
                except socket_error:
                    pass
                return
            outsock.send(chunk)

    thread = threading.Thread(target=copy)
    thread.setDaemon(1)
    thread.start()
    return thread


class CONNECTRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    # user database.  list of strings of the form "USERNAME:PASSWORD".
    # If known_users is None, no authentication is required at all.
    known_users = ["testuser:secret"]

    def authenticate_proxy(self):
        if self.known_users is None:
            return True

        authenticated = False
        auth = self.headers.get("Proxy-Authorization")
        if auth:
            parts = auth.split()
            if len(parts) == 2:
                auth_scheme, encoded = parts
                if auth_scheme.lower() == "basic":
                    decoded = base64.b64decode(encoded)
                    if decoded in self.known_users:
                        authenticated = True

        if not authenticated:
            # we should send a Proxy-Authenticate too, but it doesn't
            # matter for this test because InteProxy doesn't look at the
            # header yet.
            self.send_response(407, "Proxy Authentication Required")
            self.send_header("Proxy-Authenticate",
                             'Basic realm="inteproxytest"')
            self.send_header("Content-Type", "text/html")
            self.send_header('Connection', 'close')
            self.end_headers()
            self.wfile.write("<html><body>"
                             "Proxy requires authentication"
                             "</body><html>")
        return authenticated


    def do_CONNECT(self):
        if not self.authenticate_proxy():
            return

        real_host, real_port = urllib.splitport(self.path)
        if real_port is not None:
            real_port = int(real_port)
        else:
            self.send_error(400,
                            "Bad Request (no port number in CONNECT method)")
            return

        self.send_response(200)
        self.end_headers()

        # origin_sock = socket.socket()
        # origin_sock.connect((real_host, real_port))
        origin_sock = connect_tcp(real_host, real_port, self.log_message)
        threads = [start_copy_thread(self.connection, origin_sock),
                   start_copy_thread(origin_sock, self.connection)]

        for thread in threads:
            thread.join()



class HTTPSServer(httpserver.HTTPServer):

    supported = ssl is not None

    def __init__(self, contents, server_address=("127.0.0.1", 0),
                 RequestHandlerClass=HTTPRequestHandler,
                 keyfile=None, certfile=None):
        self.contents = contents
        self.keyfile = keyfile
        self.certfile = certfile
        httpserver.HTTPServer.__init__(self, server_address,
                                       RequestHandlerClass)

    def get_request(self):
        sock, client_address = httpserver.HTTPServer.get_request(self)
        sslsock = ssl.wrap_socket(sock, server_side=True,
                                  certfile=self.certfile,
                                  keyfile=self.keyfile,
                                  ssl_version=ssl.PROTOCOL_SSLv23)
        return sslsock, client_address
