# Copyright (C) 2007, 2008, 2009 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Support code for the test cases"""

import os
import shutil

def create_temp_dir():
    """Create a temporary directory for the test-suite and return its name.

    The temporary directory is always called temp and is created in the
    directory where the support module is located.

    If the temp directory already exists, just return the name.
    """
    name = os.path.abspath(os.path.join(os.path.dirname(__file__), "temp"))

    # if the directory already exists, we're done
    if os.path.isdir(name):
        return name

    # create the directory
    os.mkdir(name)
    return name


class FileTestMixin:

    """Mixin class for tests that use temporary files.

    Instances of this class create a test-specific sub-directory under
    the main test temp directory.  The name of the sub-directory is the
    return value of the test's id() method.
    """

    _test_specific_directory_created = False

    def create_test_specific_temp_dir(self):
        """Creates the test specific directory and returns its absolute name.
        When this method is called for the first time and the directory
        exists, if is removed, so that a specific test instance always
        starts with an empty directory.
        """
        dirname = os.path.join(create_temp_dir(), self.id())
        if not self._test_specific_directory_created:
            if os.path.islink(dirname) or os.path.isfile(dirname):
                os.remove(dirname)
            elif os.path.isdir(dirname):
                shutil.rmtree(dirname)
            os.mkdir(dirname)
            self._test_specific_directory_created = True
        return dirname

    def temp_file_name(self, basename):
        """Returns the full name of the file named basename in the temp. dir.
        """
        return os.path.join(self.create_test_specific_temp_dir(), basename)

    def create_temp_file(self, basename, contents, permissions=None):
        """Creates a file in the temp directory with the given contents.
        The optional parameter permissions should either be None (the
        default) or an int specifying the file permissions (same format
        as the second parameter of os.chmod).  The method returns the
        absolute name of the created file.
        """
        filename = self.temp_file_name(basename)
        file = open(filename, "w")
        file.write(contents)
        file.close()
        if permissions is not None:
            os.chmod(filename, permissions)
        return filename

    def check_file_contents(self, filename, contents):
        """check the contents of a file"""
        self.assertEquals(open(filename).read(), contents)


class AttributeTestMixin:

    def check_attributes(self, obj, **attrs):
        for key, value in attrs.items():
            self.assertEquals(getattr(obj, key), value)
